Unit Equation;

Interface

Uses Classes,IntvDis,Base;

Type
  TOpType      = (otUnary,otBinary);
  TOperator    = (noAdd,noSubtract,noMultiply,noDivide,noModulo, { Numeric }
                  boAnd,boOr,boXor,boNot,                        { Boolean }
                  roEqual,roUnequal,roLess,roGreater,            { Relational }
                  roLessEqual,roGreaterEqual,
                  woAnd,woOr,woXor);                             { Bitwise }
  TVariable    = (vaP1X,vaP1Y,vaP2X,vaP2Y,vaP3X,vaP3Y,vaP4X,vaP4Y,
                  vaP5X,vaP5Y,vaP6X,vaP6Y,vaP7X,vaP7Y,vaP8X,vaP8Y,
                  vaReg0,vaReg1,vaReg2,vaReg3,vaReg4,vaReg5,vaReg6,vaReg7,
                  vaRegFlags);
  TItemType    = (itNumericOp,itBooleanOp,itRelationalOp,itBitwiseOp,
                  itNumeric,itBoolean,itVariable,itMemory,itOpenPar,
                  itClosePar,itMemRead,itMemWrite,itNone);

  TItemTypeSet = Set Of TItemType;
  TOperatorSet = Set Of TOperator;

  TItemValue = Record
   Case TItemType Of
     itNumericOp,
     itBooleanOp,
  itRelationalOp,
     itBitwiseOp: (Operator : TOperator);
       itNumeric: (Number   : LongInt);
       itBoolean: (Bool     : Boolean);
      itVariable: (Variable : TVariable);
        itMemory: (Memory   : Word);
       itOpenPar,
      itClosePar,
          itNone: (Dummy    : Word);
   End; { Case }

  TItem = Class
    ItemType : TItemType;
    Value    : TItemValue;
    Constructor Init(IType: TItemType);
    Constructor InitItem(Item: TItem);
    Constructor InitOperator(IType: TItemType; IValue: TOperator);
    Constructor InitNumeric(IValue: LongInt);
    Constructor InitBoolean(IValue: Boolean);
    Constructor InitVariable(IValue: TVariable);
    Constructor InitMemory(IValue: Word);
    Constructor InitMemRead(IValue: Word);
    Constructor InitMemWrite(IValue: Word);
    Destructor  Destroy; Override;
  End;

Function FixNum(St: String): String;
Function Evaluate(Items: TStringList;
                  Const Buf: CP1600Mem;
                  Const Regs: CP1600Regs;
                  Const Flags: Word;
                  Const SpriteX,SpriteY: TSpriteLoc;
                  Const ReadMem,WriteMem: Boolean;
                  Const RWMemLoc: Word;
                  Var Error: Boolean): Boolean;
Function PostFix(St: String): TStringList;
Function ValHex(St: String): LongInt;

Implementation

Uses HexWrite,SysUtils;

Function ValHex(St: String): LongInt;
Var
  Initial : LongInt;
  I       : Integer;
  Value   : LongInt;

Begin
  If St = '' Then
  Begin
    ValHex := -1;
    Exit;
  End;
  Value   := 0;
  St      := UpperCase(St);
  I       := (Length(St) - 1) * 4;
  Initial := 1 Shl I;
  For I := 1 To Length(St) Do
  Begin
    If St[I] In ['0'..'9'] Then Inc(Value,Initial * (Ord(St[I]) - Ord('0')))
     Else
      If St[I] In ['A'..'F'] Then Inc(Value,Initial * (Ord(St[I]) - Ord('A') + 10))
       Else
       Begin
         ValHex := -1;
         Exit;
       End;
    Initial := Initial Shr 4;
  End; { For I }
  ValHex := Value;
End; { ValHex }

Function FixNum(St: String): String;
Begin
  While (St <> '') And (St[1]          = ' ') Do Delete(St,1,1);
  While (St <> '') And (St[Length(St)] = ' ') Do Delete(St,Length(St),1);
  If St = '' Then St := '0';
  FixNum := St;
End; { FixNum }

Constructor TItem.Init(IType: TItemType);
Begin
  Inherited Create;
  ItemType    := IType;
  Value.Dummy := 0;
End; { TItem.Init }

Constructor TItem.InitItem(Item: TItem);
Begin
  Inherited Create;
  ItemType := Item.ItemType;
  Value    := Item.Value
End; { TItem.InitItem }

Constructor TItem.InitOperator(IType: TItemType; IValue: TOperator);
Begin
  Inherited Create;
  ItemType       := IType;
  Value.Operator := IValue;
End; { TItem.InitOperator }

Constructor TItem.InitNumeric(IValue: LongInt);
Begin
  Inherited Create;
  ItemType     := itNumeric;
  Value.Number := IValue;
End; { TItem.InitNumeric }

Constructor TItem.InitBoolean(IValue: Boolean);
Begin
  Inherited Create;
  ItemType   := itBoolean;
  Value.Bool := IValue;
End; { TItem.InitBoolean }

Constructor TItem.InitVariable(IValue: TVariable);
Begin
  Inherited Create;
  ItemType       := itVariable;
  Value.Variable := IValue;
End; { TItem.InitVariable }

Constructor TItem.InitMemory(IValue: Word);
Begin
  Inherited Create;
  ItemType     := itMemory;
  Value.Memory := IValue;
End; { TItem.InitMemory }

Constructor TItem.InitMemRead(IValue: Word);
Begin
  Inherited Create;
  ItemType     := itMemRead;
  Value.Memory := IValue;
End; { TItem.InitMemRead }

Constructor TItem.InitMemWrite(IValue: Word);
Begin
  Inherited Create;
  ItemType     := itMemWrite;
  Value.Memory := IValue;
End; { TItem.InitMemWrite }

Destructor TItem.Destroy;
Begin
  Inherited;
End; { TItem.Destroy }

Function PostFix(St: String): TStringList;
Var
  St1       : String;
  I,J       : Integer;
  Stack     : TStringList;
  Items     : TStringList;
  Final     : TStringList;
  Finished  : Boolean;
  Error     : Boolean;
  Ok        : Boolean;
  ItemType  : TItemType;
  L         : LongInt;
  Token     : Integer;
  Item      : TItem;

  Procedure PopStack;
  Var Item: TItem;
  Begin
    If Stack.Count > 0 Then
    Begin
      Item := TItem(Stack.Objects[Stack.Count - 1]);
      Final.AddObject('',TItem.InitItem(Item));
      Stack.Objects[Stack.Count - 1].Free;
      Stack.Delete(Stack.Count - 1);
    End;
  End; { PopStack }

  Procedure PushStack;
  Var Item: TItem;
  Begin
    If Items.Count > 0 Then
    Begin
      Item := TItem(Items.Objects[0]);
      Stack.AddObject('',TItem.InitItem(Item));
    End;
  End; { PushStack }

  Procedure InsertParentheses(IType: TItemType; OType: TOpType;
                              Op: TOperatorSet; Bounds: TItemTypeSet);
  Var
    OpPos    : Integer;
    I        : Integer;
    Level    : Integer;
    Finished : Boolean;
    Item     : TItem;
    Item1    : TItem;

  Begin
    I      := 0;
    While I < Items.Count Do
    Begin
      Item := TItem(Items.Objects[I]);

      { Detect the target operator }

      If (Item.ItemType       = IType) And
         (Item.Value.Operator In Op)   Then
      Begin

        { If this is a binary operator, backtrack to find the first operand }

        If OType = otBinary Then
        Begin
          OpPos    := I;
          Level    := 0;
          Finished := False;
          Repeat
            If OpPos > 0 Then
            Begin
              Dec(OpPos);
              Item1 := TItem(Items.Objects[OpPos]);
              If Item1.ItemType = itClosePar Then Inc(Level);
              If Item1.ItemType = itOpenPar  Then Dec(Level);
              If (Level = 0) And
                 (Item1.ItemType In (Bounds + [itOpenPar])) Then
               Finished := True;
            End;
          Until (OPPos <= 0) Or Finished;
          If OpPos < I Then
          Begin
            Items.InsertObject(OpPos,'',TItem.Init(itOpenPar));
            Inc(I);
          End;
        End
        Else
        Begin
          Items.InsertObject(I,'',TItem.Init(itOpenPar));
          Inc(I);
        End;

        { Move forward to find the next operand }

        Level    := 0;
        OpPos    := I;
        Finished := False;
        Repeat
          If OpPos < (Items.Count - 1) Then
          Begin
            Inc(OpPos);
            Item1 := TItem(Items.Objects[OpPos]);
            If Item1.ItemType = itOpenPar  Then Inc(Level);
            If Item1.ItemType = itClosePar Then Dec(Level);
            If (Level = 0) And
               (Item1.ItemType In (Bounds + [itClosePar])) Then
             Finished := True;
          End;
        Until (OpPos = Items.Count - 1) Or Finished;
        If OpPos > I Then
         Items.InsertObject(OpPos + 1,'',TItem.Init(itClosePar));
        Inc(I);
      End
      Else Inc(I);
    End; { While }
  End; { InsertParentheses }

  Procedure InsNumPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itNumericOp,OType,Op,[itNumeric,itVariable,itMemory]);
  End; { InsNumPar }

  Procedure InsBoolPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itBooleanOp,OType,Op,[itBoolean]);
  End; { InsBoolPar }

  Procedure InsRelBoolPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itRelationalOp,OType,Op,[itBoolean]);
  End; { InsRelBoolPar }

  Procedure InsRelNumPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itRelationalOp,OType,Op,[itNumeric,itVariable,itMemory]);
  End; { InsRelNumPar }

  Procedure InsBitwPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itBitwiseOp,OType,Op,[itNumeric,itVariable,itMemory]);
  End; { InsBitwPar }

Begin

  { Check for mismatched parentheses }

  I     := 1;
  J     := 0;
  Error := False;
  While (I <= Length(St)) And Not Error Do
  Begin
    If St[I] = '(' Then Inc(J);
    If St[I] = ')' Then Dec(J);
    If J < 0 Then Error := True;
    Inc(I);
  End; { While }
  If J <> 0 Then Error := True;
  If Error Then
  Begin
    PostFix := Nil;
    Exit;
  End;

  { Initialize variables }

  Stack := TStringList.Create;
  Items := TStringList.Create;

  { Remove all spaces }

  I := Pos(' ',St);
  While I <> 0 Do
  Begin
    Delete(St,I,1);
    I := Pos(' ',St);
  End; { While }

  { Convert to uppercase }

  St := UpperCase(St);

  { Tokenize the string }

  I := 1;
  ItemType := itNone;
  Error    := False;
  While (I <= Length(St)) And Not Error Do
  Begin
    Token := 1;       { Default token size }

    { Detect operators (do larger ones first!) }

    If (Copy(St,I,2) = '&&') Then
    Begin
      Items.AddObject('',TItem.InitOperator(itBooleanOp,boAnd));
      Token := 2;
    End;
    If (Copy(St,I,2) = '||') Then
    Begin
      Items.AddObject('',TItem.InitOperator(itBooleanOp,boOr));
      Token := 2;
    End;
    If (Copy(St,I,2) = '^^') Then
    Begin
      Items.AddObject('',TItem.InitOperator(itBooleanOp,boXOr));
      Token := 2;
    End;
    If (Copy(St,I,2) = '<=') Then
    Begin
      Items.AddObject('',TItem.InitOperator(itRelationalOp,roLessEqual));
      Token := 2;
    End;
    If (Copy(St,I,2) = '>=') Then
    Begin
      Items.AddObject('',TItem.InitOperator(itRelationalOp,roGreaterEqual));
      Token := 2;
    End;
    If (Copy(St,I,2) = '<>') Then
    Begin
      Items.AddObject('',TItem.InitOperator(itRelationalOp,roUnequal));
      Token := 2;
    End;
    If (Copy(St,I,2) = '!=') Then
    Begin
      Items.AddObject('',TItem.InitOperator(itRelationalOp,roUnequal));
      Token := 2;
    End;
    If (Copy(St,I,2) = '==') Then
    Begin
      Items.AddObject('',TItem.InitOperator(itRelationalOp,roEqual));
      Token := 2;
    End;
    If Token = 1 Then
    Begin
      If St[I] = '~' Then Items.AddObject('',TItem.InitOperator(itBooleanOp,boNot));
      If St[I] = '!' Then Items.AddObject('',TItem.InitOperator(itBooleanOp,boNot));
      If St[I] = '=' Then Items.AddObject('',TItem.InitOperator(itRelationalOp,roEqual));
      If St[I] = '<' Then Items.AddObject('',TItem.InitOperator(itRelationalOp,roLess));
      If St[I] = '>' Then Items.AddObject('',TItem.InitOperator(itRelationalOp,roGreater));
      If St[I] = '&' Then Items.AddObject('',TItem.InitOperator(itBitwiseOp,woAnd));
      If St[I] = '|' Then Items.AddObject('',TItem.InitOperator(itBitwiseOp,woOr));
      If St[I] = '^' Then Items.AddObject('',TItem.InitOperator(itBitwiseOp,woXor));
      If St[I] = '%' Then Items.AddObject('',TItem.InitOperator(itNumericOp,noModulo));
      If St[I] = '*' Then Items.AddObject('',TItem.InitOperator(itNumericOp,noMultiply));
      If St[I] = '/' Then Items.AddObject('',TItem.InitOperator(itNumericOp,noDivide));
      If (St[I] = '+') And (ItemType In [itNumeric,itMemory,itVariable,itClosePar]) Then
       Items.AddObject('',TItem.InitOperator(itNumericOp,noAdd));
      If (St[I] = '-') And (ItemType In [itNumeric,itMemory,itVariable,itClosePar]) Then
       Items.AddObject('',TItem.InitOperator(itNumericOp,noSubtract));

      { Detect integers (decimal and hecadecimal) }

      If (St[I] In ['0'..'9']) Or
         ((St[I] In ['-','+']) And
          Not (ItemType In [itNumeric,itMemory,itVariable,itClosePar])) Then
      Begin
        J        := I + 1;
        Finished := False;
        While (J <= Length(St)) And Not Finished Do
        Begin
          If Not (St[J] In ['0'..'9','A'..'F'])
           Then Finished := True
           Else Inc(J);
        End; { While }

        { Detect hexadecimal numbers }

        If (J <= Length(St)) And (St[J] = 'H') Then Inc(J);

        { Get the number's value if possible }

        Token  := J - I;
        St1    := Copy(St,I,Token);
        If St1[1] = '+' Then Delete(St1,1,1);
        If St1[Length(St1)] <> 'H' Then Val(St1,L,J)
        Else
        Begin
          St1 := Copy(St1,1,Length(St1) - 1);  { Strip off the "h" }
          L := 0;
          If St1[1] = '-' Then Error := True Else L := ValHex(St1);
          If L < 0 Then Error := True;
        End;
        If Not Error Then Items.AddObject('',TItem.InitNumeric(L));
      End;

      { Detect numeric variables }

      If (Copy(St,I,3) = 'S1X') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP1X));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S1Y') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP1Y));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S2X') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP2X));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S2Y') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP2Y));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S3X') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP3X));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S3Y') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP3Y));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S4X') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP4X));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S4Y') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP4Y));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S5X') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP5X));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S5Y') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP5Y));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S6X') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP6X));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S6Y') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP6Y));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S7X') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP7X));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S7Y') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP7Y));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S8X') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP8X));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S8Y') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaP8Y));
        Token := 3;
      End;

      If (Copy(St,I,2) = 'R0') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaReg0));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R1') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaReg1));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R2') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaReg2));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R3') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaReg3));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R4') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaReg4));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R5') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaReg5));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R6') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaReg6));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R7') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaReg7));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'SW') Then
      Begin
        Items.AddObject('',TItem.InitVariable(vaRegFlags));
        Token := 2;
      End;
      If Token = 1 Then
      Begin

        { Detect reading from a memory location }

        If Copy(St,I,2) = 'R[' Then
        Begin
          J        := I + 2;
          Ok       := True;
          Finished := False;
          While (J <= Length(St)) And Ok And Not Finished Do
          Begin
            If Not (St[J] In ['0'..'9','A'..'F',']']) Then Ok := False;
            If St[J] = ']' Then
            Begin
              Finished := True;
              If J = I + 2 Then Ok := False;  { Detect "[]" }
            End;
            Inc(J);
          End; { While }
          If Finished And Ok Then
          Begin
            Token := J - I;
            L     := ValHex(Copy(St,I + 2,Token - 3)); { Strip off brackets }
            If L >= 0
             Then Items.AddObject('',TItem.InitMemRead(L))
             Else Error := True;
          End
          Else Error := True;
        End;

        { Detect writing to a memory location }

        If Copy(St,I,2) = 'W[' Then
        Begin
          J        := I + 2;
          Ok       := True;
          Finished := False;
          While (J <= Length(St)) And Ok And Not Finished Do
          Begin
            If Not (St[J] In ['0'..'9','A'..'F',']']) Then Ok := False;
            If St[J] = ']' Then
            Begin
              Finished := True;
              If J = I + 2 Then Ok := False;  { Detect "[]" }
            End;
            Inc(J);
          End; { While }
          If Finished And Ok Then
          Begin
            Token := J - I;
            L     := ValHex(Copy(St,I + 2,Token - 3)); { Strip off brackets }
            If L >= 0
             Then Items.AddObject('',TItem.InitMemWrite(L))
             Else Error := True;
          End
          Else Error := True;
        End;

        { Detect memory references (always hexadecimal) }

        If St[I] = '[' Then
        Begin
          J        := I + 1;
          Ok       := True;
          Finished := False;
          While (J <= Length(St)) And Ok And Not Finished Do
          Begin
            If Not (St[J] In ['0'..'9','A'..'F',']']) Then Ok := False;
            If St[J] = ']' Then
            Begin
              Finished := True;
              If J = I + 1 Then Ok := False;  { Detect "[]" }
            End;
            Inc(J);
          End; { While }
          If Finished And Ok Then
          Begin
            Token := J - I;
            L     := ValHex(Copy(St,I + 1,Token - 2)); { Strip off brackets }
            If L >= 0
             Then Items.AddObject('',TItem.InitMemory(L))
             Else Error := True;
          End
          Else Error := True;
        End;

        { Detect parentheses }

        If St[I] = '(' Then Items.AddObject('',TItem.Init(itOpenPar));
        If St[I] = ')' Then Items.AddObject('',TItem.Init(itClosePar));
      End;
    End;

    { Remember the type of whatever we added for next time }

    If Items.Count > 0 Then
     ItemType := TItem(Items.Objects[Items.Count - 1]).ItemType;

    { Move to the next token }

    Inc(I,Token);
  End; { While }

  { If the string was tokenized okay, continue }

  If Not Error Then
  Begin

    { Insert multiplier operations where multiplication is implicit }

    Final    := TStringList.Create;
    I        := 0;
    ItemType := itNone;
    While I < Items.Count Do
    Begin
      Item := TItem(Items.Objects[I]);

      { Insert multipliers where needed }

      If (Item.ItemType In [itNumeric,itVariable,itMemory,itOpenPar]) And
         (ItemType      In [itNumeric,itVariable,itMemory,itClosePar]) Then
      Begin
        Items.InsertObject(I,'',TItem.InitOperator(itNumericOp,noMultiply));
        ItemType := itNumericOp;
      End
      Else ItemType := Item.ItemType; { Remember the type for next time }
      Inc(I);
    End; { While }

    { Add sets of parentheses for operator precedence }

    InsBoolPar(otUnary,[boNot]);
    InsNumPar(otBinary,[noMultiply,noDivide,noModulo]);
    InsBoolPar(otBinary,[boAnd]);
    InsBitwPar(otBinary,[woAnd]);
    InsNumPar(otBinary,[noAdd,noSubtract]);
    InsBoolPar(otBinary,[boOr,boXor]);
    InsBitwPar(otBinary,[woOr,woXor]);
    InsRelNumPar(otBinary,[roEqual,roUnequal,roLess,roGreater]);
    InsRelBoolPar(otBinary,[roEqual,roUnequal]);
    InsRelNumPar(otBinary,[roLessEqual,roGreaterEqual]);

    { Convert to postfix }

    While Items.Count > 0 Do
    Begin

      ItemType := TItem(Items.Objects[0]).ItemType;
      Case ItemType Of
        itClosePar: PopStack;  { Pop an operator off the stack }

       itNumericOp,
       itBooleanOp,
    itRelationalOp,
       itBitwiseOp: PushStack; { Push any operators onto the stack }

         itNumeric,
        itVariable,
          itMemory,
         itMemRead,
        itMemWrite,
         itBoolean: Final.AddObject('',TItem.InitItem(TItem(Items.Objects[0])));
      End; { Case }

      { Remove the item from the infix list }

      Items.Objects[0].Free;
      Items.Delete(0);
    End; { While }

    { Pop any operators off the stack }

    While Stack.Count > 0 Do PopStack;
  End
  Else Final := Nil;
  Items.Free;
  Stack.Free;
  PostFix := Final;
End; { PostFix }

Function Evaluate(Items: TStringList;
                  Const Buf: CP1600Mem;
                  Const Regs: CP1600Regs;
                  Const Flags: Word;
                  Const SpriteX,SpriteY: TSpriteLoc;
                  Const ReadMem,WriteMem: Boolean;
                  Const RWMemLoc: Word;
                  Var Error: Boolean): Boolean;
Var
  Func  : TItem;
  I     : Integer;
  Stack : TStringList;  { TCollections take WAY too long to do this }
  E1    : TItem;        { on every Atari instruction                }
  E2    : TItem;
  Item  : TItem;
  Count : Integer;
  Value : TItemValue;

  Procedure PopStack(Var Item: TItem);
  Begin
    If Not Error Then
    Begin
      If Stack.Count > 0 Then
      Begin
        Item.ItemType := TItem(Stack.Objects[Stack.Count - 1]).ItemType;
        Item.Value    := TItem(Stack.Objects[Stack.Count - 1]).Value;
        Stack.Objects[Stack.Count - 1].Free;
        Stack.Delete(Stack.Count - 1);
      End
      Else Error := True;
    End;
  End; { PopStack }

Begin
  Func  := TItem.Create;
  Error := False;
  E1    := TItem.Create;
  E2    := TItem.Create;
  If Items <> Nil Then
  Begin
    Count := Items.Count;

    { There should never be more than this on the stack }

    I     := 0;
    Error := False;
    Stack := TStringList.Create;
    While (I < Count) And Not Error Do
    Begin
      Item             := TItem(Items.Objects[I]);
      Value            := Item.Value;
      Func.ItemType    := itNone;
      Func.Value.Dummy := 0;

      { Handle operators }

      Case Item.ItemType Of
        itNumericOp,
        itBooleanOp,
        itRelationalOp,
        itBitwiseOp:
        Begin
          Case Value.Operator Of
            noAdd:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E1.Value.Number + E2.Value.Number;
              End
              Else Error := True;
            End;
            noSubtract:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number - E1.Value.Number;
              End
              Else Error := True;
            End;
            noMultiply:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E1.Value.Number * E2.Value.Number;
              End
              Else Error := True;
            End;
            noDivide:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) And
                 (E1.Value.Number <> 0)    Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number Div E1.Value.Number;
              End
              Else Error := True;
            End;
            noModulo:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) And
                 (E1.Value.Number <> 0)    Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number Mod E1.Value.Number;
              End
              Else Error := True;
            End;
            boAnd:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := E2.Value.Bool And E1.Value.Bool;
              End
              Else Error := True;
            End;
            boOr:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := E2.Value.Bool Or E1.Value.Bool;
              End
              Else Error := True;
            End;
            boXor:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := E2.Value.Bool Xor E1.Value.Bool;
              End
              Else Error := True;
            End;
            boNot:
            Begin
              PopStack(E1);
              If E1.ItemType = itBoolean Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := Not E1.Value.Bool;
              End
              Else If E1.ItemType = itNumeric Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := Not E1.Value.Number;
              End
              Else Error := True;
            End;
            roEqual:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Bool = E1.Value.Bool);
              End
              Else If (E1.ItemType = itNumeric) And
                      (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number = E1.Value.Number);
              End
              Else Error := True;
            End;
            roUnequal:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Bool <> E1.Value.Bool);
              End
              Else If (E1.ItemType = itNumeric) And
                      (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number <> E1.Value.Number);
              End
              Else Error := True;
            End;
            roLess:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number < E1.Value.Number);
              End
              Else Error := True;
            End;
            roGreater:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number > E1.Value.Number);
              End
              Else Error := True;
            End;
            roLessEqual:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number <= E1.Value.Number);
              End
              Else Error := True;
            End;
            roGreaterEqual:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number >= E1.Value.Number);
              End
              Else Error := True;
            End;
            woAnd:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number And E1.Value.Number;
              End
              Else Error := True;
            End;
            woOr:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number Or E1.Value.Number;
              End
              Else Error := True;
            End;
            woXor:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number Xor E1.Value.Number;
              End
              Else Error := True;
            End;
          Else
            Error := True;
          End; { Case }
        End;
        itNumeric,
        itBoolean: Func.InitItem(Item);
        itVariable:
        Begin
          Func.ItemType := itNumeric;
          Case Value.Variable Of
          vaP1X: Func.Value.Number := SpriteX[0];
          vaP1Y: Func.Value.Number := SpriteY[0];
          vaP2X: Func.Value.Number := SpriteX[1];
          vaP2Y: Func.Value.Number := SpriteY[1];
          vaP3X: Func.Value.Number := SpriteX[2];
          vaP3Y: Func.Value.Number := SpriteY[2];
          vaP4X: Func.Value.Number := SpriteX[3];
          vaP4Y: Func.Value.Number := SpriteY[3];
          vaP5X: Func.Value.Number := SpriteX[4];
          vaP5Y: Func.Value.Number := SpriteY[4];
          vaP6X: Func.Value.Number := SpriteX[5];
          vaP6Y: Func.Value.Number := SpriteY[5];
          vaP7X: Func.Value.Number := SpriteX[6];
          vaP7Y: Func.Value.Number := SpriteY[6];
          vaP8X: Func.Value.Number := SpriteX[7];
          vaP8Y: Func.Value.Number := SpriteY[7];
         vaReg0: Func.Value.Number := Regs[0];
         vaReg1: Func.Value.Number := Regs[1];
         vaReg2: Func.Value.Number := Regs[2];
         vaReg3: Func.Value.Number := Regs[3];
         vaReg4: Func.Value.Number := Regs[4];
         vaReg5: Func.Value.Number := Regs[5];
         vaReg6: Func.Value.Number := Regs[6];
         vaReg7: Func.Value.Number := Regs[7];
     vaRegFlags: Func.Value.Number := Flags;
          Else
            Error := True;
          End; { Case }
        End;
        itMemory:
        Begin
          Func.ItemType := itNumeric;
          If Value.Memory < $2000 Then
           Func.Value.Number := Buf[Value.Memory]
          Else Error := True;
        End;
        itMemRead:
        Begin
          Func.ItemType   := itBoolean;
          Func.Value.Bool := ReadMem And (RWMemLoc = Value.Memory)
        End;
        itMemWrite:
        Begin
          Func.ItemType   := itBoolean;
          Func.Value.Bool := WriteMem And (RWMemLoc = Value.Memory)
        End;
      End; { Case }
      Inc(I);
      If Not Error Then Stack.AddObject('',TItem.InitItem(Func));
    End; { While }
    Stack.Free;
    If Func.ItemType <> itBoolean Then Error := True;
  End
  Else Error := True;
  If Not Error Then Evaluate := Func.Value.Bool Else Evaluate := False;
  Func.Free;
  E1.Free;
  E2.Free;
End; { Evaluate }

End.

// ----------------------------------------------------------------------
// IntvWin - Mattel Intellivision Emulator
// Copyright (C) 2000 John Dullea
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// ----------------------------------------------------------------------
