Unit Dis6502;

Interface

Uses Profile;

Type
  TDisArray   = Packed Array[0..62 * 1024 - 1] Of Boolean;
  TGameROM    = Packed Array[0..65535] Of Byte;
  TASeg       = Packed Array[0..65535] Of Byte;

Const
  ListingFileName : String = 'LISTING.ASM';
  BreakInstruction = 2;

  Timer_And : Packed Array[$0..$1F] Of Byte =
   ($80,$81,$82,$83,$84,$85,$84,$85,$80,$81,$82,$83,$84,$85,$84,$85,
    $94,$95,$96,$97,$94,$95,$96,$97,$94,$95,$96,$97,$94,$95,$96,$97);

  Bytes : Packed Array[0..255] Of Byte =
   (1,2,1,2,2,2,2,2, 1,2,1,2,3,3,3,3,  { 00-0F }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { 10-1F }
    3,2,1,2,2,2,2,2, 1,2,1,2,3,3,3,3,  { 20-2F }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { 30-3F }
    1,2,1,2,2,2,2,2, 1,2,1,2,3,3,3,3,  { 40-4F }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { 50-5F }
    1,2,1,2,2,2,2,2, 1,2,1,2,3,3,3,3,  { 60-6F }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { 70-7F }
    2,2,2,2,2,2,2,2, 1,1,1,2,3,3,3,3,  { 80-8F }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { 90-9F }
    2,2,2,1,2,2,2,2, 1,2,1,2,3,3,3,3,  { A0-AF }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { B0-BF }
    2,2,2,2,2,2,2,2, 1,2,1,2,3,3,3,3,  { C0-CF }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { D0-DF }
    2,2,2,2,2,2,2,2, 1,2,1,2,3,3,3,3,  { E0-EF }
    2,2,1,2,2,2,2,2, 1,3,1,3,1,3,3,3); { F0-FF }

  Mnem : Packed Array[0..255] Of String[3] =
   ('BRK','ORA','   ','slo','skb','ORA','ASL','slo',   { 00-07 }
    'PHP','ORA','ASL','anc','skw','ORA','ASL','slo',   { 08-0F }
    'BPL','ORA','   ','slo','skb','ORA','ASL','slo',   { 10-17 }
    'CLC','ORA','NOP','slo','skw','ORA','ASL','slo',   { 18-1F }
    'JSR','AND','   ','rla','BIT','AND','ROL','rla',   { 20-27 }
    'PLP','AND','ROL','anc','BIT','AND','ROL','rla',   { 28-2F }
    'BMI','AND','   ','rla','skb','AND','ROL','rla',   { 30-37 }
    'SEC','AND','NOP','rla','skw','AND','ROL','rla',   { 38-3F }
    'RTI','EOR','   ','sre','skb','EOR','LSR','sre',   { 40-47 }
    'PHA','EOR','LSR','asr','JMP','EOR','LSR','sre',   { 48-4F }
    'BVC','EOR','   ','sre','skb','EOR','LSR','sre',   { 50-57 }
    'CLI','EOR','NOP','sre','skw','EOR','LSR','sre',   { 58-5F }
    'RTS','ADC','   ','rra','skb','ADC','ROR','rra',   { 60-67 }
    'PLA','ADC','ROR','arr','JMP','ADC','ROR','rra',   { 68-6F }
    'BVS','ADC','   ','rra','skb','ADC','ROR','rra',   { 70-77 }
    'SEI','ADC','NOP','rra','skw','ADC','ROR','rra',   { 78-7F }
    'skb','STA','skb','sax','STY','STA','STX','sax',   { 80-87 }
    'DEY','NOP','TXA','ane','STY','STA','STX','sax',   { 88-8F }
    'BCC','STA','   ','sha','STY','STA','STX','sax',   { 90-97 }
    'TYA','STA','TXS','shs','shy','STA','shx','sha',   { 98-9F }
    'LDY','LDA','LDX','lax','LDY','LDA','LDX','lax',   { A0-A7 }
    'TAY','LDA','TAX','lxa','LDY','LDA','LDX','lax',   { A8-AF }
    'BCS','LDA','   ','lax','LDY','LDA','LDX','lax',   { B0-B7 }
    'CLV','LDA','TSX','las','LDY','LDA','LDX','lax',   { B8-BF }
    'CPY','CMP','skb','dcp','CPY','CMP','DEC','dcp',   { C0-C7 }
    'INY','CMP','DEX','sbx','CPY','CMP','DEC','dcp',   { C8-CF }
    'BNE','CMP','   ','dcp','skb','CMP','DEC','dcp',   { D0-D7 }
    'CLD','CMP','NOP','dcp','skw','CMP','DEC','dcp',   { D8-DF }
    'CPX','SBC','skb','isb','CPX','SBC','INC','isb',   { E0-E7 }
    'INX','SBC','NOP','SBC','CPX','SBC','INC','isb',   { E8-EF }
    'BEQ','SBC','   ','isb','skb','SBC','INC','isb',   { F0-F7 }
    'SED','SBC','NOP','isb','skw','SBC','INC','isb');  { F8-FF }

  ATyp : Packed Array[0..255] Of String[3] =
   ('IMP','I.X','   ','I.X','IMP','Z.P','Z.P','Z.P',   { 00-07 }
    'IMP','IMM','ACC','IMM','IMP','ABS','ABS','ABS',   { 08-0F }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { 10-17 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.X','A.X',   { 18-1F }
    'ABS','I.X','   ','I.X','Z.P','Z.P','Z.P','Z.P',   { 20-27 }
    'IMP','IMM','ACC','IMM','ABS','ABS','ABS','ABS',   { 28-2F }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { 30-37 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.X','A.X',   { 38-3F }
    'IMP','I.X','   ','I.X','IMP','Z.P','Z.P','Z.P',   { 40-47 }
    'IMP','IMM','ACC','IMM','ABS','ABS','ABS','ABS',   { 48-4F }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { 50-57 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.X','A.X',   { 58-5F }
    'IMP','I.X','   ','I.X','IMP','Z.P','Z.P','Z.P',   { 60-67 }
    'IMP','IMM','ACC','IMM','IND','ABS','ABS','ABS',   { 68-6F }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { 70-77 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.X','A.X',   { 78-7F }
    'IMP','I.X','IMP','I.X','Z.P','Z.P','Z.P','Z.P',   { 80-87 }
    'IMP','IMP','IMP','IMM','ABS','ABS','ABS','ABS',   { 88-8F }
    'REL','I.Y','   ','I.Y','Z.X','Z.X','Z.Y','Z.Y',   { 90-97 }
    'IMP','A.Y','IMP','IMP','A.X','A.X','A.Y','A.Y',   { 98-9F }
    'IMM','I.X','IMM','I.X','Z.P','Z.P','Z.P','Z.P',   { A0-A7 }
    'IMP','IMM','IMP','I+A','ABS','ABS','ABS','ABS',   { A8-AF }
    'REL','I.Y','   ','I.Y','Z.X','Z.X','Z.Y','Z.Y',   { B0-B7 }
    'IMP','A.Y','IMP','ASY','A.X','A.X','A.Y','A.Y',   { B8-BF }
    'IMM','I.X','IMP','I.X','Z.P','Z.P','Z.P','Z.P',   { C0-C7 }
    'IMP','IMM','IMP','IMM','ABS','ABS','ABS','ABS',   { C8-CF }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { D0-D7 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.X','A.X',   { D8-DF }
    'IMM','I.X','IMP','I.X','Z.P','Z.P','Z.P','Z.P',   { E0-E7 }
    'IMP','IMM','IMP','IMM','ABS','ABS','ABS','ABS',   { E8-EF }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { F0-F7 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.Y','A.X');  { F8-FF }

Var
  Executed : TDisArray;
  Branched : TDisArray;

Function Mem6502W(Addr: Word): Word;
Function ROMBackupW(Addr: Word): Word;

Implementation

Uses HexWrite,Base;

Function Mem6502W(Addr: Word): Word;
Var L,H: Word;
Begin
  L := Mem6502[Addr];
  H := Mem6502[Addr + 1];
  Mem6502W := (H Shl 8) + L;
End; // Mem6502W

Function ROMBackupW(Addr: Word): Word;
Var L,H: Word;
Begin
  L := ROMBackup[Addr];
  H := ROMBackup[Addr + 1];
  ROMBackupW := L + (H Shl 8);
End; { ROMBackupW }

End.

// ----------------------------------------------------------------------
// IntvWin - Mattel Intellivision Emulator
// Copyright (C) 2000 John Dullea
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// ----------------------------------------------------------------------
